<?php declare(strict_types=1);

namespace Convo\Core\Adapters\Webchat;

use Psr\Http\Server\RequestHandlerInterface;
use Convo\Core\Rest\RestSystemUser;

class WebchatRestHandler implements RequestHandlerInterface
{
	/**
	 * @var \Convo\Core\Factory\ConvoServiceFactory
	 */
	private $_convoServiceFactory;

	/**
	 * @var \Convo\Core\IServiceDataProvider
	 */
	private $_convoServiceDataProvider;

	/**
	 * @var \Convo\Core\Params\IServiceParamsFactory
	 */
	private $_convoServiceParamsFactory;

	/**
	 * @var \Convo\Core\Factory\PlatformRequestFactory
	 */
	private $_platformRequestFactory;

	/**
	 * @var \Convo\Core\Util\IHttpFactory
	 */
	private $_httpFactory;

	/**
	 * @var \Psr\Log\LoggerInterface
	 */
	private $_logger;

	public function __construct( $logger, $httpFactory, $serviceFactory, $serviceDataProvider, $serviceParamsFactory, $platformRequestFactory)
	{
		$this->_logger						=	$logger;
		$this->_httpFactory					=	$httpFactory;
		$this->_convoServiceFactory			= 	$serviceFactory;
		$this->_convoServiceDataProvider	= 	$serviceDataProvider;
		$this->_convoServiceParamsFactory	= 	$serviceParamsFactory;
		$this->_platformRequestFactory  	=	$platformRequestFactory;
	}

	public function handle(\Psr\Http\Message\ServerRequestInterface $request): \Psr\Http\Message\ResponseInterface
	{
		$info	=	new \Convo\Core\Rest\RequestInfo( $request);

		$this->_logger->debug( 'Got info ['.$info.']');

		if ( $info->post() && $route = $info->route( 'service-run/webchat/{variant}/{serviceId}'))
		{
			$variant = $route->get('variant');
			$serviceId = $route->get('serviceId');

			$this->_logger->debug("Executing [$serviceId][$variant]");

			return $this->_handleWebchatPathServiceIdPost($request, $variant, $serviceId);
		}

		throw new \Convo\Core\Rest\NotFoundException( 'Could not map ['.$info.']');
	}

	private function _handleWebchatPathServiceIdPost(\Psr\Http\Message\ServerRequestInterface $request, $variant, $serviceId)
	{
		$owner		=	new RestSystemUser();
		$json		=	$request->getParsedBody();

		$text		=	$json['text'] ?? null;
		$is_init	=	$json['lunch'] ?? false;
		$device_id	=	$json['device_id'];


// 		try {
// 			$meta	=	$this->_convoServiceDataProvider->getServiceMeta( $owner, $serviceId);
// 		} catch ( \Convo\Core\ComponentNotFoundException $e) {
// 			throw new \Convo\Core\Rest\NotFoundException( 'Service meta ['.$serviceId.'] not found', 0, $e);
// 		}

		try {
			$version_id			=	$this->_convoServiceFactory->getVariantVersion( $owner, $serviceId, 'convo_chat', $variant);
		} catch ( \Convo\Core\ComponentNotFoundException $e) {
			throw new \Convo\Core\Rest\NotFoundException( 'Service variant ['.$serviceId.']['.$variant.'] not found', 0, $e);
		}

		$this->_logger->debug( 'Got variant version ['.$variant.']['.$version_id.']');

		if ( empty( $version_id)) {
			throw new \Convo\Core\Rest\InvalidRequestException( 'Service ['.$serviceId.'] not published yet');
		}

		try {
			$platform_config	=	$this->_convoServiceDataProvider->getServicePlatformConfig( $owner, $serviceId, $version_id);
		} catch ( \Convo\Core\ComponentNotFoundException $e) {
			throw new \Convo\Core\Rest\NotFoundException( 'Service platform config ['.$serviceId.']['.$version_id.'] not found', 0, $e);
		}

		$this->_logger->debug( 'Got config ['.print_r( $platform_config, true).']');

		if ( !isset( $platform_config['convo_chat'])) {
		    throw new \Convo\Core\Rest\InvalidRequestException( 'Service ['.$serviceId.'] version ['.$version_id.'] is not enabled for platform ['.'convo_chat'.']');
		}

		$delegate_nlp		=	$platform_config['convo_chat']['delegateNlp'] ?? null;

		$service			=	$this->_convoServiceFactory->getService( $owner, $serviceId, $version_id);

		$text_request		=	new \Convo\Core\Adapters\Webchat\DefaultTextCommandRequest(
			$serviceId, $device_id, $device_id, $device_id, $text, $is_init, false /* <- temp */, $delegate_nlp);
		$text_response		=	new \Convo\Core\Adapters\Webchat\DefaultTextCommandResponse();
		$text_response->setLogger($this->_logger);

		if ( $delegate_nlp) {
			// 		    TODO: load & use service owner account
			// 		    $service_meta     =   $this->_convoServiceDataProvider->getServiceMeta( $user, $service_id);
			// 		    $owner            =   $service_meta['owner'];
            $text_request     =   $this->_platformRequestFactory->toIntentRequest($text_request, $owner, $serviceId, $delegate_nlp);
		}

		$service->run( $this->_convoServiceParamsFactory, $text_request, $text_response);

		$data	=	array(
				'service_state' => $service->getServiceState(),
		);

		$data		=	array_merge( $data, $text_response->getPlatformResponse());
		return $this->_httpFactory->buildResponse( $data);
	}

	// UTIL
	public function __toString()
	{
		return get_class( $this).'[]';
	}
}
