<?php


namespace Convo\Pckg\Filesystem;


use Convo\Core\Params\IServiceParamsScope;
use Convo\Core\Util\Mp3File;
use Convo\Core\Workflow\AbstractBasicComponent;
use Convo\Core\Workflow\IMediaSourceContext;
use wapmorgan\Mp3Info\Mp3Info;

class FilesystemMediaContext extends AbstractBasicComponent implements IMediaSourceContext
{
    private $_id;

    /** @var string */
    private $_folderPath = "";

    /** @var string */
    private $_basePath = "";

    /** @var boolean */
    private $_shouldMovePonter = true;

    /**
     *
     * @var \Psr\Log\LoggerInterface
     */
    protected $_logger;

    public function __construct($properties)
    {
        parent::__construct($properties);
        $this->_id			    =	$properties['id'];
        $this->_folderPath		=	$properties['mp3SourcePath'];
        $this->_basePath		=	$properties['mp3SourceBasePath'];
    }

    /**
     * @inheritDoc
     */
    public function list(): iterable
    {
        if ($this->_shouldMovePonter) {
            $this->_setCurrentSongIndex(0);
        }
        $songs = array_slice(scandir($this->_folderPath), 2);
        $availableSongs = [];
        foreach ($songs as $song) {
            $songFile =  $this->_folderPath . DIRECTORY_SEPARATOR . $song;
            array_push($availableSongs, $songFile);
        }
        $this->_logger->debug("All available songs [" . print_r($availableSongs, true) . "] from [" . $this->_folderPath . "]");
        $this->_getServiceParams()->setServiceParam('available_songs', $availableSongs);
        return $availableSongs;
    }

    /**
     * @inheritDoc
     */
    public function find($searchQuery): iterable
    {
        $availableSongList = $this->list();
        $filteredSongsList = [];
        $searchTextSimilarityPercentage = 75;

        foreach ($availableSongList as $songFile) {
            $songData = $this->_prepareSong($songFile);

            if ($songData->isMetaDataAvailable()) {
                if (isset($searchQuery['Artist']) && !isset($searchQuery['Song'])) {
                    $percent = 100;
                    similar_text(strtolower($songData->getArtist()), strtolower($searchQuery['Artist']), $percent);

                    if ($percent < $searchTextSimilarityPercentage) {
                        continue;
                    }
                } else if (isset($searchQuery['Genre'])) {
                    $percent = 100;
                    similar_text(strtolower($songData->getGenre()), strtolower($searchQuery['Genre']), $percent);

                    if ($percent < $searchTextSimilarityPercentage) {
                        continue;
                    }
                } else if (isset($searchQuery['Artist']) && isset($searchQuery['Song'])) {
                    $percentArtist = 100;
                    similar_text(strtolower($songData->getArtist()), strtolower($searchQuery['Artist']), $percentArtist);
                    $percentSongTitle = 100;
                    similar_text(strtolower($songData->getSongTitle()), strtolower($searchQuery['Song']), $percentSongTitle);
                    $percentAvg = ($percentArtist + $percentSongTitle) / 2;

                    if ($percentAvg < $searchTextSimilarityPercentage) {
                        continue;
                    }
                }
            } else {
                if (isset($searchQuery['Artist']) && !isset($searchQuery['Song'])) {
                    $this->_logger->debug("Search by file name for artist");
                    $artistFromFileName = explode("-", $songData->getFileName())[0];
                    $percent = 100;
                    similar_text(strtolower(trim($artistFromFileName)), strtolower($searchQuery['Artist']), $percent);

                    if ($percent < $searchTextSimilarityPercentage) {
                        continue;
                    }
                } else if (isset($searchQuery['Artist']) && isset($searchQuery['Song'])) {
                    $this->_logger->debug("Search by file name for song by artist");
                    $artistFromFileName = explode("-", $songData->getFileName())[0];
                    $songFromFileName = explode("-", $songData->getFileName())[1];

                    $percentArtist = 100;
                    similar_text(strtolower(trim($artistFromFileName)), strtolower($searchQuery['Artist']), $percentArtist);
                    $percentSongTitle = 100;
                    similar_text(strtolower(trim($songFromFileName)), strtolower($searchQuery['Song']), $percentSongTitle);
                    $percentAvg = ($percentArtist + $percentSongTitle) / 2;

                    if ($percentAvg < $searchTextSimilarityPercentage) {
                        continue;
                    }
                }
            }
            array_push($filteredSongsList, $songFile);
        }

        $this->_getServiceParams()->setServiceParam('available_songs', $filteredSongsList);
        return $filteredSongsList;
    }

    /**
     * @inheritDoc
     */
    public function current(): Mp3File
    {
        $song = new Mp3File('', '', [], '');
        if (isset($this->_getAvailableSongs()[$this->_getCurrentSongIndex()])) {
            $song = $this->_getAvailableSongs()[$this->_getCurrentSongIndex()];
        }
        return $this->_prepareSong($song);
    }

    /**
     * @inheritDoc
     */
    public function next(): Mp3File
    {
        $song = new Mp3File('', '', [], '');
        $songIndex = $this->_getNextSongIndex();

        if (!empty($this->_getAvailableSongs()[$songIndex])) {
            if ($this->_shouldMovePonter) {
                $this->_setCurrentSongIndex($songIndex);
            }
            $songFile = $this->_getAvailableSongs()[$songIndex];
            $song = $this->_prepareSong($songFile);
        }

        return $song;
    }

    /**
     * @inheritDoc
     */
    public function previous(): Mp3File
    {
        $song = new Mp3File('', '', [], '');
        $songIndex = $this->_getPreviousSongIndex();
        if (!empty($this->_getAvailableSongs()[$songIndex])) {
            if ($this->_shouldMovePonter) {
                $this->_setCurrentSongIndex($songIndex);
            }
            $songFile = $this->_getAvailableSongs()[$songIndex];
            $song = $this->_prepareSong($songFile);
        }

        return $song;
    }

    public function first(): Mp3File
    {
        if ($this->_shouldMovePonter) {
            $this->_setCurrentSongIndex(0);
        }
        return $this->_prepareSong($this->_getAvailableSongs()[0]);
    }

    public function last(): Mp3File
    {
        $songIndex = count($this->_getAvailableSongs()) - 1;
        if ($this->_shouldMovePonter) {
            $this->_setCurrentSongIndex($songIndex);
        }
        return $this->_prepareSong($this->_getAvailableSongs()[$songIndex]);
    }

    /**
     * @inheritDoc
     */
    public function setOffset($offset)
    {
        $this->_getServiceParams()->setServiceParam("current_song_offset", $offset);
    }

    /**
     * @inheritDoc
     */
    public function getOffset(): int
    {
        $currentSongOffsetFromInstallation = $this->_getServiceParams()->getServiceParam("current_song_offset");

        if (is_numeric($currentSongOffsetFromInstallation)) {
            return $currentSongOffsetFromInstallation;
        }

        return 0;
    }

    public function movePointerTo($index)
    {
        $this->_setCurrentSongIndex($index);
    }

    public function getPointerPosition(): int
    {
        return $this->_getCurrentSongIndex();
    }

    public function setLoopStatus($loopStatus)
    {
        $this->_getServiceParams()->setServiceParam('loop_status', true);
    }

    public function getLoopStatus(): bool
    {
        return $this->_getServiceParams()->getServiceParam('loop_status') ? $this->_getServiceParams()->getServiceParam('loop_status') : false;
    }

    public function setShouldMovePointer($shouldMovePointer = true)
    {
        $this->_shouldMovePonter = $shouldMovePointer;
    }

    /**
     * @return string[]
     */
    private function _getAvailableSongs()
    {
        return $this->_getServiceParams()->getServiceParam("available_songs");
    }

    private function _setCurrentSongIndex($currentSongIndex) {
        $this->_getServiceParams()->setServiceParam("current_song_index", $currentSongIndex);
    }

    private function _getCurrentSongIndex() {
        $currentSongIndex = 0;
        $currentSongIndexFromInstallation = $this->_getServiceParams()->getServiceParam("current_song_index");

        if (is_numeric($currentSongIndexFromInstallation)) {
            $currentSongIndex = $currentSongIndexFromInstallation;
        }

        $this->_setCurrentSongIndex($currentSongIndex);
        return $currentSongIndex;
    }

    private function _getNextSongIndex() {
        return $this->_getCurrentSongIndex() + 1;
    }

    private function _getPreviousSongIndex() {
        return $this->_getCurrentSongIndex() - 1;
    }

    private function _prepareSong($songFile) {
        $fileMetaData = [];

        $this->_logger->debug("Song file path [" . $songFile . "]");
        $dirName = basename(dirname($songFile));
        $songUrl = $this->_basePath . $dirName ."/" . basename($songFile);

        try {
            $audio = new Mp3Info($songFile, true);
            $fileMetaData = $audio->tags;
        } catch (\Exception $e) {
            $this->_logger->warning($e->getMessage());
            return new Mp3File(basename($songFile), $songUrl, $fileMetaData, basename(dirname($songFile)));
        }

        $this->_logger->debug("Song file path [" . $songFile . "]");

        return new Mp3File(basename($songFile), $songUrl, $fileMetaData, basename(dirname($songFile)));
    }

    private function _getServiceParams() {
        return $this->getService()->getServiceParams(IServiceParamsScope::SCOPE_TYPE_INSTALLATION);
    }

    /**
     * @inheritDoc
     */
    public function init()
    {
        // TODO: init folder path and base path.
    }

    /**
     * @inheritDoc
     */
    public function getId()
    {
        return $this->_id;
    }

    /**
     * @inheritDoc
     */
    public function getComponent()
    {
        $this->init();
        return $this;
    }
}
